//Biblioteca de calculo dos ganhos PID (Kp, Ki, Kd)
//Compativel com ESP32, Arduino e placas compatíveis  
//método Ziegler-Nichols
//Desenvolvida por Opller

#include <OP_AutoTune.h>

Autotune::Autotune() {
}


void Autotune::config(int target, unsigned long interval, int min, int max, int tuneCycles, int potencia){
  targetInputValue = target;
  loopInterval = interval;
  percentual_potencia = potencia;
  minOutput = min;
  maxOutput = max*percentual_potencia/100;
  cycles = tuneCycles;
}

// modo Ziegler-Nichols
void Autotune::setZNMode(ZNMode zn) {
  znMode = zn;
}

// Inicialização das variáveis
void Autotune::start(unsigned long us) {
  i = 0; // Cycle counter
  output = true; // Current output state
  outputValue = maxOutput;
  t1 = t2 = us; // Times used for calculating period
  millisseconds = tHigh = tLow = 0; // More time variables
  max = -1000000000000; // Max input
  min = 1000000000000; // Min input
  pAverage = iAverage = dAverage = 0;
}

//Executa a função
//enable -> habilitado = 1 / desabilitado = 0
//input -> variável a ser controlada
// us -> tempo atual

float Autotune::execute(bool enable, int input, unsigned long us) {
  
  // Explicação básica de como funciona:
  // Liga a saida do controle na potencia escolhida
  // Espera que o sistema chegue no setpoint e desliga a saida
  // Espera que o sistema abaixe do setpoint e liga a saida novamente
  // Repita o ciclo pela quantidade de vezes escolhida
  // Calcula os ganhos pelo metodo Ziegler-Nichols
  
  enable_control = enable; 
  
  if(enable_control == 1){
  millisseconds = us;
  
  max = (max > input) ? max : input;
  min = (min < input) ? min : input; 

  if (output && input > targetInputValue) {
    output = false;
    outputValue = minOutput;
    t1 = us;
    tHigh = t1 - t2;
    max = targetInputValue;
  }

  if (!output && input < targetInputValue - 20) {
    // Turn output on, record current time as t2, calculate tLow
    output = true;
    outputValue = maxOutput;
    t2 = us;
    tLow = t2 - t1;

    float ku = (4.0 * ((maxOutput - minOutput) / 2.0)) / (M_PI * (max - min) / 2.0);

    float tu = tLow + tHigh;

    float kpConstant, tiConstant, tdConstant;
    if (znMode == ZNModeBasicPID) {
      kpConstant = 0.6;
      tiConstant = 0.5;
      tdConstant = 0.125;
    } else if (znMode == ZNModeLessOvershoot) {
      kpConstant = 0.33;
      tiConstant = 0.5;
      tdConstant = 0.33;
    } else { // Default to No Overshoot mode as it is the safest
      kpConstant = 0.2;
      tiConstant = 0.5;
      tdConstant = 0.33;
    }

    kp = kpConstant * ku;
    ki = (kp / (tiConstant * tu)) * loopInterval;
    kd = (tdConstant * kp * tu) / loopInterval;

    //if (i > 1) {
      pAverage += kp;
      iAverage += ki;
      dAverage += kd;
    //}
    
    i++;
    min = targetInputValue;
  }

  if (i >= cycles) {
    output = false;
    outputValue = minOutput;
    kp = pAverage / (i);
    ki = iAverage / (i);
    kd = dAverage / (i);
  }
  } else {
    output = false;
    outputValue = 0;
  }
  
  return outputValue;
}

//Funçoes para ler os ganhos calculados
float Autotune::getKp() { return kp; };
float Autotune::getKi() { return ki; };
float Autotune::getKd() { return kd; };

//Retorna true se o numero de ciclos encerrou
bool Autotune::isFinished() {
  return (i >= cycles);
}

//Retorna o numero de ciclos
int Autotune::getCycle() {
  return i;
}
